from pypif.obj.common import Property, Scalar

from .base import DFTParser, Value_if_true, InvalidIngesterException
import os
from pypif.obj.common.value import Value
from dftparse.pwscf.stdout_parser import PwscfStdOutputParser
from ase import Atoms


class PwscfParser(DFTParser):
    '''
    Parser for PWSCF calculations
    '''

    def __init__(self, files):
        super(PwscfParser, self).__init__(files)
        self.settings = {}
        parser = PwscfStdOutputParser()

        # Look for appropriate files
        self.inputf = self.outputf = None
        for f in self._files:
            try:
                if self._get_line('Program PWSCF', f, return_string=False):
                    if self.outputf is not None:
                        raise InvalidIngesterException('More than one output file!')
                    self.outputf = f
                elif self._get_line('&control', f, return_string=False, case_sens=False):
                    if self.inputf is not None:
                        raise InvalidIngesterException('More than one input file')
                    self.inputf = f
            except UnicodeDecodeError as e:
                pass

        if self.inputf is None:
            raise InvalidIngesterException('Failed to find input file')
        if self.outputf is None:
            raise InvalidIngesterException('Failed to find output file')

        # Read in the settings
        with open(self.outputf, "r") as f:
            for line in parser.parse(f.readlines()):
                self.settings.update(line)

    def get_result_functions(self):
        base_results = super(PwscfParser, self).get_result_functions()
        base_results["One-electron energy contribution"] = "get_one_electron_energy_contribution"
        base_results["Hartree energy contribution"] = "get_hartree_energy_contribution"
        base_results["Exchange-correlation energy contribution"] = "get_xc_energy_contribution"
        base_results["Ewald energy contribution"] = "get_ewald_energy_contribution"
        return base_results

    def get_name(self): return "PWSCF"

    def _get_key_with_units(self, key):
        if key not in self.settings:
            return None
        return Property(scalars=[Scalar(value=self.settings[key])], units=self.settings["{} units".format(key)])

    def _get_line(self, search_string, search_file, return_string=True, case_sens=True):
        '''Return the first line containing a set of strings in a file.

        If return_string is False, we just return whether such a line
        was found. If case_sens is False, the search is case
        insensitive.

        '''
        if os.path.isfile(search_file):
            # if single search string
            if type(search_string) == type(''): search_string = [search_string]
            # if case insensitive, convert everything to lowercase
            if not case_sens: search_string = [i.lower() for i in search_string]
            with open(search_file) as fp:
                # search for the strings line by line
                for line in fp:
                    query_line = line if case_sens else line.lower()
                    if all([i in query_line for i in search_string]):
                        return line if return_string else True
                if return_string:
                    raise Exception('%s not found in %s'%(' & '.join(search_string), search_file))
                else: return False
        else: raise Exception('%s file does not exist'%search_file)

    def get_version_number(self):
        '''Determine the version number from the output'''
        return self.settings["version"]

    def get_xc_functional(self):
        '''Determine the xc functional from the output'''
        return Value(scalars=[Scalar(value=" ".join(self.settings["exchange-correlation"]))])

    def get_cutoff_energy(self):
        '''Determine the cutoff energy from the output'''
        return Value(
            scalars=[Scalar(value=self.settings["kinetic-energy cutoff"])],
            units=self.settings['kinetic-energy cutoff units']
        )

    def get_total_energy(self):
        '''Determine the total energy from the output'''
        return self._get_key_with_units("total energy")

    @Value_if_true
    def is_relaxed(self):
        '''Determine if relaxation run from the output'''
        return self._get_line('Geometry Optimization', self.outputf, return_string=False)

    def _is_converged(self):
        '''Determine if calculation converged; for a relaxation (static) run
        we look for ionic (electronic) convergence in the output'''
        if self.is_relaxed():
            # relaxation run case
            return self._get_line(['End of', 'Geometry Optimization'], self.outputf, return_string=False)
        else:
            # static run case
            return self._get_line('convergence has been achieved', self.outputf, return_string=False)

    def get_KPPRA(self):
        '''Determine the no. of k-points in the BZ (from the input) times the
        no. of atoms (from the output)'''
        # Find the no. of k-points
        fp = open(self.inputf).readlines()
        for l,ll in enumerate(fp):
            if "K_POINTS" in ll:
                # determine the type of input
                if len(ll.split()) > 1:
                    if "gamma" in ll.split()[1].lower():
                        ktype = 'gamma'
                    elif "automatic" in ll.split()[1].lower():
                        ktype = 'automatic'
                    else:
                        ktype = ''
                else: ktype = ''
                if ktype == 'gamma':
                    # gamma point:
                    # K_POINTS {gamma}
                    nk = 1
                elif ktype == 'automatic':
                    # automatic:
                    # K_POINTS automatic
                    #  12 12 1 0 0 0
                    line = [int(i) for i in fp[l+1].split()[0:3]]
                    nk = line[0]*line[1]*line[2]
                else:
                    # manual:
                    # K_POINTS
                    #  3
                    #  0.125  0.125  0.0  1.0
                    #  0.125  0.375  0.0  2.0
                    #  0.375  0.375  0.0  1.0
                    nk = 0
                    for k in range(int(fp[l+1].split()[0])):
                        nk += int(float(fp[l+2+k].split()[3]))
                # Find the no. of atoms
                natoms = int(self._get_line('number of atoms/cell', self.outputf).split()[4])
                return Value(scalars=[Scalar(value=nk*natoms)])
        fp.close()
        raise Exception('%s not found in %s'%('KPOINTS',self.inputf))

    @Value_if_true
    def uses_SOC(self):
        '''Looks for line with "with spin-orbit" in the output'''
        return self._get_line('with spin-orbit', self.outputf, return_string=False)

    def get_pp_name(self):
        '''Determine the pseudopotential names from the output'''
        ppnames = []
        # Find the number of atom types
        natomtypes = int(self._get_line('number of atomic types', self.outputf).split()[5])
        # Find the pseudopotential names
        with open(self.outputf) as fp:
            for line in fp:
                if "PseudoPot. #" in line:
                    ppnames.append(Scalar(value=next(fp).split('/')[-1].rstrip()))
                    if len(ppnames) == natomtypes:
                        return Value(scalars=ppnames)
            raise Exception('Could not find %i pseudopotential names'%natomtypes)

    def get_U_settings(self):
        '''Determine the DFT+U type and parameters from the output'''
        with open(self.outputf) as fp:
            for line in fp:
                if "LDA+U calculation" in line:
                    U_param = {}
                    U_param['Type'] = line.split()[0]
                    U_param['Values'] = {}
                    # look through next several lines
                    for nl in range(15):
                        line2 = next(fp).split()
                        if len(line2) > 1 and line2[0] == "atomic":
                            pass # column titles
                        elif len(line2) == 6:
                            U_param['Values'][line2[0]] = {}
                            U_param['Values'][line2[0]]['L'] = float(line2[1])
                            U_param['Values'][line2[0]]['U'] = float(line2[2])
                            U_param['Values'][line2[0]]['J'] = float(line2[4])
                        else: break # end of data block
                    return Value(**U_param)
            return None

    def get_vdW_settings(self):
        '''Determine the vdW type if using vdW xc functional or correction
        scheme from the input otherwise'''
        xc = self.get_xc_functional().scalars[0].value
        if 'vdw' in xc.lower(): # vdW xc functional
            return Value(scalars=[Scalar(value=xc)])
        else:
            # look for vdw_corr in input
            vdW_dict = {'xdm':'Becke-Johnson XDM', 'ts':
                        'Tkatchenko-Scheffler', 'ts-vdw':
                        'Tkatchenko-Scheffler',
                        'tkatchenko-scheffler':
                        'Tkatchenko-Scheffler', 'grimme-d2': 'Grimme D2', 'dft-d': 'Grimme D2'}
            if self._get_line('vdw_corr', self.inputf, return_string=False, case_sens=False):
                line = self._get_line('vdw_corr', self.inputf, return_string=True, case_sens=False)
                vdwkey = str(line.split('=')[-1].replace("'", "").replace(',', '').lower().rstrip())
                return Value(scalars=[Scalar(value=vdW_dict[vdwkey])])
            return None

    def get_pressure(self):
        '''Determine the pressure from the output'''
        return self._get_key_with_units("pressure")

    def get_stresses(self):
        '''Determine the stress tensor from the output'''
        if "stress" not in self.settings:
            return None
        wrapped = [[Scalar(value=x) for x in y] for y in self.settings["stress"]]
        return Property(matrices=[wrapped], units=self.settings["stress units"])

    def get_output_structure(self):
        '''Determine the structure from the output'''
        bohr_to_angstrom = 0.529177249

        # determine the number of atoms
        natoms = int(float(self._get_line('number of atoms/cell', self.outputf).split('=')[-1]))

        # determine the initial lattice parameter
        alat = float(self._get_line('lattice parameter (alat)', self.outputf).split('=')[-1].split()[0])

        # find the initial unit cell
        unit_cell = []
        with open(self.outputf, 'r') as fp:
            for line in fp:
                if "crystal axes:" in line:
                    for i in range(3):
                        unit_cell.append([float(j)*alat*bohr_to_angstrom for j in next(fp).split('(')[-1].split(')')[0].split()])
                    break
            if len(unit_cell) == 0: raise Exception('Cannot find the initial unit cell')

        # find the initial atomic coordinates
        coords = [] ; atom_symbols = []
        with open(self.outputf, 'r') as fp:
            for line in fp:
                if "site n." in line and "atom" in line and "positions" in line and "alat units" in line:
                    for i in range(natoms):
                        coordline = next(fp)
                        atom_symbols.append(''.join([i for i in coordline.split()[1] if not i.isdigit()]))
                        coord_conv_factor = alat*bohr_to_angstrom
                        coords.append([float(j)*coord_conv_factor for j in coordline.rstrip().split('=')[-1].split('(')[-1].split(')')[0].split()])
                    break
            if len(coords) == 0: raise Exception('Cannot find the initial atomic coordinates')

        if type(self.is_relaxed()) == type(None):
            # static run: create, populate, and return the initial structure
            structure = Atoms(symbols=atom_symbols, cell=unit_cell, pbc=True)
            structure.set_positions(coords)
            return structure
        else:
            # relaxation run: update with the final structure
            with open(self.outputf) as fp:
                for line in fp:
                    if "Begin final coordinates" in line:
                        if 'new unit-cell volume' in next(fp):
                            # unit cell allowed to change
                            next(fp) # blank line
                            # get the final unit cell
                            unit_cell = []
                            cellheader = next(fp)
                            if 'bohr' in cellheader.lower():
                                cell_conv_factor = bohr_to_angstrom
                            elif 'angstrom' in cellheader.lower():
                                cell_conv_factor = 1.0
                            else:
                                alat = float(cellheader.split('alat=')[-1].replace(')', ''))
                                cell_conv_factor = alat*bohr_to_angstrom
                            for i in range(3):
                                unit_cell.append([float(j)*cell_conv_factor for j in next(fp).split()])
                            next(fp) # blank line

                        # get the final atomic coordinates
                        coordtype = next(fp).split()[-1].replace('(', '').replace(')', '')
                        if coordtype == 'bohr':
                            coord_conv_factor = bohr_to_angstrom
                        elif coordtype == 'angstrom' or coordtype == 'crystal':
                            coord_conv_factor = 1.0
                        else:
                            coord_conv_factor = alat*bohr_to_angstrom
                        coords = [] # reinitialize the coords
                        for i in range(natoms):
                            coordline = next(fp).split()
                            coords.append([float(j)*coord_conv_factor for j in coordline[1:4]])

                        # create, populate, and return the final structure
                        structure = Atoms(symbols=atom_symbols, cell=unit_cell, pbc=True)
                        if coordtype == 'crystal':
                            structure.set_scaled_positions(coords) # direct coord
                        else:
                            structure.set_positions(coords) # cartesian coord
                        return structure
                raise Exception('Cannot find the final coordinates')

    def get_dos(self):
        '''Find the total DOS shifted by the Fermi energy'''
        # find the dos file
        fildos = ''
        for f in self._files:
            with open(f, 'r') as fp:
                first_line = next(fp)
                if "E (eV)" in first_line and "Int dos(E)" in first_line:
                    fildos = f
                    ndoscol = len(next(fp).split())-2 # number of spin channels
                    fp.close()
                    break
                fp.close()
        if not fildos:
            return None # cannot find DOS

        # get the Fermi energy
        line = self._get_line('the Fermi energy is', self.outputf)
        efermi = float(line.split('is')[-1].split()[0])

        # grab the DOS
        energy = [] ; dos = []
        fp = open(fildos, 'r')
        next(fp) # comment line
        for line in fp:
            ls = line.split()
            energy.append(Scalar(value=float(ls[0])-efermi))
            dos.append(Scalar(value=sum([float(i) for i in ls[1:1+ndoscol]])))
        return Property(scalars=dos, units='number of states per unit cell', conditions=Value(name='energy', scalars=energy, units='eV'))

    def get_forces(self):
        if "forces" not in self.settings:
            return None
        wrapped = [[Scalar(value=x) for x in y] for y in self.settings['forces']]
        return Property(vectors=wrapped, units=self.settings['force units'])

    def get_total_force(self):
        if "total force" not in self.settings:
            return None
        return Property(scalars=[Scalar(value=self.settings['total force'])], units=self.settings['force units'])

    def get_outcar(self):
        return None

    def get_incar(self):
        return None

    def get_poscar(self):
        return None

    def get_band_gap(self):
        '''Compute the band gap from the DOS'''
        dosdata = self.get_dos()
        if type(dosdata) == type(None):
            return None # cannot find DOS
        else:
            energy = dosdata.conditions.scalars
            dos = dosdata.scalars
            step_size = energy[1].value - energy[0].value
            not_found = True ; l = 0 ; bot = 10**3 ; top = -10**3
            while not_found and l < len(dos):
                # iterate through the data
                e = float(energy[l].value)
                dens = float(dos[l].value)
                # note: dos already shifted by efermi
                if e < 0 and dens > 1e-3:
                    bot = e
                elif e > 0 and dens > 1e-3:
                    top = e
                    not_found = False
                l += 1
            if top < bot:
                raise Exception('Algorithm failed to find the band gap')
            elif top - bot < step_size*2:
                return Property(scalars=[Scalar(value=0)], units='eV')
            else:
                bandgap = float(top-bot)
                return Property(scalars=[Scalar(value=round(bandgap,3))], units='eV')

    def get_one_electron_energy_contribution(self):
        return self._get_key_with_units("one-electron energy contribution")

    def get_hartree_energy_contribution(self):
        return self._get_key_with_units("hartree energy contribution")

    def get_xc_energy_contribution(self):
        return self._get_key_with_units("xc energy contribution")

    def get_ewald_energy_contribution(self):
        return self._get_key_with_units("ewald energy contribution")



from .base import DFTParser, InvalidIngesterException
import os
import glob
from ase.calculators.abinit import Abinit
from pypif.obj.common import Value, Property, Scalar


class AbinitParser(DFTParser):
    '''
    Parser for ABINIT calculations
    '''
    _label = None

    def __init__(self, files):
        # Check whether any file has as name ABINIT in the file in the first two lines
        super(AbinitParser, self).__init__(files)
        for f in files:
            try:
                with open(f, 'r') as fp:
                    for line in [fp.readline(), fp.readline()]:
                        if "ABINIT" in line:
                            is_abinit = True
            except:
                continue
        if not is_abinit:
            raise InvalidIngesterException('No Abinit files found')
    
    def get_name(self): return "ABINIT"

    def _get_label(self):
        '''Find the label for the output files 
         for this calculation
        '''
        if self._label is None:
            foundfiles = False
            for f in self._files:
                if ".files" in f: 
                    foundfiles = True
                    self._label = f.split(".")[0]
                    with open(self._label + '.files', 'r') as fp:
                        line = fp.readline().split()[0]
                        if line != self._label + ".in":
                           fp.close()
                           raise Exception('first line must be label.in')
                        line = fp.readline().split()[0]
                        if line != self._label + ".txt":
                           fp.close()
                           raise Exception('second line must be label.txt')
                        line = fp.readline().split()[0]
                        if line != self._label + "i":
                           fp.close()
                           raise Exception('third line must be labeli')
                        line = fp.readline().split()[0]
                        if line != self._label + "o":
                           fp.close()
                           raise Exception('fourth line must be labelo')
                        fp.close()
            if foundfiles:
                return self._label
            else:
                raise Exception('label.files not found')
                                
#ASE format
#        (self.prefix + '.in') # input
#        (self.prefix + '.txt')# output
#        (self.prefix + 'i')   # input
#        (self.prefix + 'o')   # output
                
        else:
            return self._label

    def get_cutoff_energy(self):
        if not self._label:
            self._get_label()
        # Open up the label.txt file
        fp = open(os.path.join(self._directory, self._label + '.out'), 'r')
        foundecho = False 
        # Look for ecut after the occurence of "echo values of preprocessed input variables"
        for line in fp:
            if "echo values of preprocessed input variables" in line:
                foundecho = True
            if "ecut" in line and foundecho:
                words = line.split()
                return Value(scalars=[Scalar(value=float(words[1]))], units=words[2])
        # Error handling: ecut not found
        raise Exception('ecut not found')



from pypif.obj import Property, Scalar

from .base import DFTParser, Value_if_true, InvalidIngesterException
import os
import re
from ase.io.vasp import read_vasp_out
from pypif.obj import Value, FileReference
from dftparse.vasp.outcar_parser import OutcarParser
from dftparse.vasp.eigenval_parser import EigenvalParser


class VaspParser(DFTParser):
    '''
    Parser for VASP calculations
    '''

    def __init__(self, files):
        super(VaspParser, self).__init__(files)

        # Find the outcar file
        def _find_file(name):
            """Find a filename that contains a certain string"""
            name = name.upper()

            my_file = None
            for f in self._files:
                if os.path.basename(f).upper().startswith(name):
                    if my_file is not None:
                        raise InvalidIngesterException('Found more than one {} file'.format(name))
                    my_file = f
            return my_file
        self.outcar = _find_file('OUTCAR')
        if self.outcar is None:
            raise InvalidIngesterException('OUTCAR not found!')

        # Find the DOSCAR, EIGENVAL, and INCAR files
        #   None of these are required so we do not throw exceptions
        self.incar = _find_file('INCAR')
        self.poscar = _find_file('POSCAR')
        self.doscar = _find_file('DOSCAR')
        self.eignval = _find_file('EIGNVAL')

    def get_name(self): return "VASP"
        
    def get_output_structure(self):
        self.atoms = read_vasp_out(self.outcar)
        return self.atoms

    def get_outcar(self):
        raw_path = self.outcar
        if raw_path[0:2] == "./":
            raw_path = raw_path[2:]
        return Property(files=[FileReference(
            relative_path=raw_path
        )])

    def get_incar(self):
        if self.incar is None: return None
        raw_path = self.incar
        if raw_path[0:2] == "./":
            raw_path = raw_path[2:]
        return Value(files=[FileReference(
            relative_path=raw_path
        )])

    def get_poscar(self):
        if self.poscar is None: return None
        raw_path = self.poscar
        if raw_path[0:2] == "./":
            raw_path = raw_path[2:]
        return Value(files=[FileReference(
            relative_path=raw_path
        )])

    def get_cutoff_energy(self):
        # Open up the OUTCAR
        with open(self.outcar, 'r') as fp:
         # Look for ENCUT
            for line in fp:
                if "ENCUT" in line:
                    words = line.split()
                    return Value(scalars=[Scalar(value=float(words[2]))], units=words[3])
                
        # Error handling: ENCUT not found
        raise Exception('ENCUT not found')

    @Value_if_true
    def uses_SOC(self):
        # Open up the OUTCAR
        with open(self.outcar) as fp:
        
            #look for LSORBIT
            for line in fp:
                if "LSORBIT" in line:
                    words = line.split()
                    return words[2] == 'T'
        
        # Error handling: LSORBIT not found
        raise Exception('LSORBIT not found')
        
    @Value_if_true
    def is_relaxed(self):
        # Open up the OUTCAR
        with open(self.outcar) as fp:
        
            #  Look for NSW
            for line in fp:
                if "NSW" in line:
                    words = line.split()
                    return int(words[2]) != 0
                
        # Error handling: NSW not found
        raise Exception('NSW not found')
        
    def get_xc_functional(self):
        # Open up the OUTCAR
        with open(self.outcar) as fp:
        
            # Look for TITEL
            for line in fp:
                if "TITEL" in line:
                    words = line.split()
                    return Value(scalars=[Scalar(value=words[2])])
            
    def get_pp_name(self):
        # Open up the OUTCAR
        with open(self.outcar) as fp:
        
            #initialize empty list to store pseudopotentials
            pp = []
            # Look for TITEL
            for line in fp:
                if "TITEL" in line:
                    words = line.split()
                    pp.append(words[3])
            return Value(vectors=[[Scalar(value=x) for x in pp]])

    def get_KPPRA(self):
        # Open up the OUTCAR
        with open(self.outcar) as fp:
            #store the number of atoms and number of irreducible K-points
            for line in fp:
                if "number of ions     NIONS =" in line:
                    words = line.split()
                    NI = int(words[11])
                elif "k-points           NKPTS =" in line:
                    words = line.split()
                    NIRK = float(words[3])
            #check if the number of k-points was reduced by VASP if so, sum all the k-points weight
            if "irreducible" in open(self.outcar).read():
                fp.seek(0)
                for line in fp:
                    #sum all the k-points weight
                    if "Coordinates               Weight" in line:
                        NK=0; counter = 0
                        for line in fp:
                            if counter == NIRK:
                                break
                            NK += float(line.split()[3])
                            counter += 1
                return Value(scalars=[Scalar(value=NI*NK)])
            #if k-points were not reduced KPPRA equals the number of atoms * number of irreducible k-points
            else:
                return Value(scalars=[Scalar(value=NI*NIRK)])

    def _is_converged(self):
        # Follows the procedure used by qmpy, but without reading the whole file into memory
        #   Source: https://github.com/wolverton-research-group/qmpy/blob/master/qmpy/analysis/vasp/calculation.py

        with open(self.outcar) as fp:
            # Part 1: Determine the NELM
            nelm = None
            for line in fp:
                if line.startswith("   NELM   ="):
                    nelm = int(line.split()[2][:-1])
                    break

            # If we don't find it, tell the user
            if nelm is None:
                raise Exception('NELM not found. Cannot tell if this result is converged')

            # Now, loop through the file. What we want to know is whether the last ionic
            #   step of this file terminates because it converges or because we hit NELM
            re_iter = re.compile('([0-9]+)\( *([0-9]+)\)')
            converged = False
            for line in fp:
                # Count the ionic steps
                if 'Iteration' in line:
                    ionic, electronic = map(int, re_iter.findall(line)[0])

                # If the loop is finished, mark the number of electronic steps
                if 'aborting loop' in line:
                    converged = electronic < nelm

            return converged

    def get_total_energy(self):
        with open(self.outcar) as fp:
            last_energy = None
            for line in fp:
                if line.startswith('  free  energy   TOTEN'):
                    last_energy = float(line.split()[4])
        if last_energy is None:
            return None
        return Property(scalars=[Scalar(value=last_energy)], units='eV')

    def get_version_number(self):
        # Open up the OUTCAR
        with open(self.outcar) as fp:
        
            #look for vasp
            for line in fp:
                if "vasp" in line:
                    words = line.split()
                    return (words[0].strip('vasp.'))
                    break
        
        # Error handling: vasp not found
        raise Exception('vasp not found')
        
    def get_U_settings(self):
        #Open up the OUTCAR
        with open(self.outcar) as fp:
            #Check if U is used
            if "LDAU" in open(self.outcar).read():
                U_param = {}
                atoms = []
                #get the list of pseupotential used
                for line in fp:
                    if "TITEL" in line:
                        atoms.append(line.split()[3])
                    #Get the U type used
                    if "LDAUTYPE" in line:
                            U_param['Type'] = int(line.split()[-1])
                atoms.reverse()
                fp.seek(0)
                #Get the L value
                U_param['Values'] = {}
                for line in fp:
                    for atom, i in zip(atoms, range(len(atoms))):
                        if "LDAUL" in line:
                                U_param['Values'][atom] = {'L': int(line.split()[-1-i])}
                fp.seek(0)
                #Get the U value            
                for line in fp:
                    for atom, i in zip(atoms, range(len(atoms))):
                        if "LDAUU" in line:
                                U_param['Values'][atom]['U'] = float(line.split()[-1-i])
                fp.seek(0)
                #Get the J value
                for line in fp:
                    for atom, i in zip(atoms, range(len(atoms))):
                        if "LDAUJ" in line:
                                U_param['Values'][atom]['J'] = float(line.split()[-1-i])
                return Value(**U_param)
            #if U is not used, return None
            else:
                return None
            
    def get_vdW_settings(self):
        #define the name of the vdW methods in function of their keyword
        vdW_dict = {'BO':'optPBE-vdW', 'MK':'optB88-vdW', 'ML':'optB86b-vdW','RE':'vdW-DF','OR':'Klimes-Bowler-Michaelides'}
        #Open up the OUTCAR
        with open(self.outcar) as fp:
            #Check if vdW is used
            if "LUSE_VDW" in open(self.outcar).read():
                #if vdW is used, get its keyword             
                for line in fp:
                    if "GGA     =" in line:
                        words = line.split()
                        return Value(scalars=[Scalar(value=vdW_dict[words[2]])])
            #if vdW is not used, return None
            else:
                return None
            
    def get_pressure(self):
        #define pressure dictionnary because since when is kB = kbar? Come on VASP people
        pressure_dict = {'kB':'kbar'}
        #Check if ISIF = 0 is used
        if "ISIF   =      0" in open(self.outcar).read():
            #if ISIF = 0 is used, print this crap
            return None
        #if ISIF is not 0 then extract pressure and units
        else:
            #scan file in reverse to have the final pressure
            for line in reversed(open(self.outcar).readlines()):
                if "external pressure" in line:
                    words = line.split()
                    return Property(scalars=[Scalar(value=float(words[3]))], units=pressure_dict[words[4]])
                    break
    
    def get_stresses(self):
        #Check if ISIF = 0 is used
        if "ISIF   =      0" in open(self.outcar).read():
            return None
        #Check if ISIF = 1 is used
        elif "ISIF   =      1" in open(self.outcar).read():
            return None
        else:
            #scan file in reverse to have the final pressure
            for line in open(self.outcar).readlines():
                if "in kB" in line:
                    words = line.split()
                    XX = float(words[2]); YY = float(words[3]); ZZ = float(words[4]); XY= float(words[5]); YZ = float(words[6]); ZX = float(words[7])
            matrix = [[XX,XY,ZX],[XY,YY,YZ],[ZX,YZ,ZZ]]
            wrapped = [[Scalar(value=x) for x in y] for y in matrix]
            return Property(matrices=[wrapped], units='kbar')

    def get_forces(self):
        self.atoms = read_vasp_out(self.outcar)
        forces_raw = self.atoms.get_calculator().results['forces'].tolist()
        forces_wrapped = [[Scalar(value=x) for x in y] for y in forces_raw]
        positions_raw = self.atoms.positions.tolist()
        positions_wrapped = [[Scalar(value=x) for x in y] for y in positions_raw]
        return Property(
            vectors=forces_wrapped,
            conditions=Value(name="positions", vectors=positions_wrapped)
        )

    @staticmethod
    def _get_bandgap_from_bands(energies, nelec):
        """Compute difference in conduction band min and valence band max"""
        nelec = int(nelec)
        valence = [x[nelec-1] for x in energies]
        conduction = [x[nelec] for x in energies]
        return max(min(conduction) - max(valence), 0.0)

    @staticmethod
    def _get_bandgap_eigenval(eigenval_fname, outcar_fname):
        """Get the bandgap from the EIGENVAL file"""
        with open(outcar_fname, "r") as f:
            parser = OutcarParser()
            nelec = next(iter(filter(lambda x: "number of electrons" in x, parser.parse(f.readlines()))))["number of electrons"]
        with open(eigenval_fname, "r") as f:
            eigenval_info = list(EigenvalParser().parse(f.readlines()))
        # spin_polarized = (2 == len(next(filter(lambda x: "kpoint" in x, eigenval_info))["occupancies"][0]))
        # if spin_polarized:
        all_energies = [zip(*x["energies"]) for x in eigenval_info if "energies" in x]
        spin_energies = zip(*all_energies)
        gaps = [VaspParser._get_bandgap_from_bands(x, nelec/2.0) for x in spin_energies]
        return min(gaps)

    @staticmethod
    def _get_bandgap_doscar(filename):
        """Get the bandgap from the DOSCAR file"""
        with open(filename) as fp:
            for i in range(6):
                l = fp.readline()
            efermi = float(l.split()[3])
            step1 = fp.readline().split()[0]
            step2 = fp.readline().split()[0]
            step_size = float(step2)-float(step1)
            not_found = True
            while not_found:
                l = fp.readline().split()
                e = float(l.pop(0))
                dens = 0.0
                for i in range(int(len(l)/2)):
                    dens += float(l[i])
                if e < efermi and dens > 1e-3:
                    bot = e
                elif e > efermi and dens > 1e-3:
                    top = e
                    not_found = False
            if top - bot < step_size*2:
                bandgap = 0.0
            else:
                bandgap = float(top - bot)

        return bandgap

    def get_band_gap(self):
        """Get the bandgap, either from the EIGENVAL or DOSCAR files"""
        if self.outcar is not None and self.eignval is not None:
            bandgap = VaspParser._get_bandgap_eigenval(self.eignval, self.outcar)
        elif self.doscar is not None:
            bandgap = VaspParser._get_bandgap_doscar(self.doscar)
        else:
            return None
        return Property(scalars=[Scalar(value=round(bandgap, 3))], units='eV')
                
    def get_dos(self):
        if self.doscar is None:
            return None
        #open DOSCAR
        with open(self.doscar) as fp:
            for i in range(6):
                l = fp.readline()
            n_step = int(l.split()[2])
            energy = []; dos = []            
            for i in range(n_step):
                l = fp.readline().split()
                e = float(l.pop(0))
                energy.append(Scalar(value=e))
                dens = 0
                for j in range(int(len(l)/2)):
                    dens += float(l[j])
                dos.append(Scalar(value=dens))

            # Convert to property
            return Property(scalars=dos, units='number of states per unit cell',
                            conditions=Value(name='energy', scalars=energy, units='eV'))

    def get_total_magnetization(self):
        if self.outcar is None:
            return None
        parser = OutcarParser()
        with open(self.outcar, "r") as fp:
            matches = list(filter(lambda x: "total magnetization" in x, parser.parse(fp.readlines())))
        if len(matches) == 0:
            return None
        total_magnetization = matches[-1]["total magnetization"]
        return Property(scalars=[Scalar(value=total_magnetization)], units="Bohr")



import os
import uuid
import tarfile
import shutil
from dfttopif.parsers import VaspParser
from dfttopif.parsers import PwscfParser
from dfttopif.parsers.base import InvalidIngesterException
from pypif.obj import *
import json


def _add_quality_report(parser, pif, inline=True):
    import tarfile

    # if the parser lacks an INCAR, return None
    if parser.incar is None:
        return None

    # Create the tar file
    tar = tarfile.open("tmp.tar", "w")
    tar.add(parser.outcar, arcname="OUTCAR")
    tar.add(parser.incar, arcname="INCAR")
    tar.close()

    import requests
    if inline:
        r = requests.post('https://calval.citrination.com/validate/json/tarfile', data=open('tmp.tar', 'rb').read())
        report = json.loads(r.json()[0])
        score = report["score"]
    else:
        r = requests.post('https://calval.citrination.com/validate/tarfile', data=open('tmp.tar', 'rb').read())
        report = r.json()[0]
        score = int(report.split('\n')[0].split()[-1]) # the score is the last token on the first line
    os.remove("tmp.tar")

    if r.status_code != requests.codes.ok:
        print("Unable to generate quality report; request returned with status {}".format(r.status_code))
        return

    if inline:
        setattr(pif, "quality_report", report)
    else:
        report_file = os.path.join(os.path.dirname(parser.outcar), "quality_report.txt")
        with open(report_file, "w") as f:
            f.write(report)
        if report_file[0:2] == "./":
            report_file = report_file[2:]
        pif.properties.append(
                Property(
                    name="quality_report",
                    scalars=[Scalar(value=score)],
                    files=[FileReference(relative_path=report_file)]
                )
            )

    return pif


def tarfile_to_pif(filename, temp_root_dir='', verbose=0):
    """
    Process a tar file that contains DFT data.

    Input:
        filename - String, Path to the file to process.
        temp_root_dir - String, Directory in which to save temporary files. Defaults to working directory.
        verbose - int, How much status messages to print

    Output:
        pif - ChemicalSystem, Results and settings of
            the DFT calculation in pif format
    """
    temp_dir = temp_root_dir + str(uuid.uuid4())
    os.makedirs(temp_dir)
    try:
        tar = tarfile.open(filename, 'r')
        tar.extractall(path=temp_dir)
        tar.close()
        for i in os.listdir(temp_dir):
            cur_dir = temp_dir + '/' + i
            if os.path.isdir(cur_dir):
                return directory_to_pif(cur_dir, verbose=verbose)
        return directory_to_pif(temp_dir, verbose=verbose)
    finally:
        shutil.rmtree(temp_dir)


def archive_to_pif(filename, verbose=0):
    """
    Given a archive file that contains output from a DFT calculation, parse the data and return a PIF object.

    Input:
        filename - String, Path to the file to process.
        verbose - int, How much status messages to print

    Output:
        pif - ChemicalSystem, Results and settings of
            the DFT calculation in pif format
    """
    if tarfile.is_tarfile(filename):
        return tarfile_to_pif(filename, verbose)
    raise Exception('Cannot process file type')


def files_to_pif(files, verbose=0, quality_report=True, inline=True):
    '''Given a directory that contains output from
    a DFT calculation, parse the data and return
    a pif object

    Input:
        files - [str] list of files from which the parser is allowed to read.
        verbose - int, How much status messages to print

    Output:
        pif - ChemicalSystem, Results and settings of
            the DFT calculation in pif format
    '''

    # Look for the first parser compatible with the directory
    found_parser = False
    for possible_parser in [PwscfParser, VaspParser]:
        try:
            parser = possible_parser(files)
            found_parser = True
            break
        except InvalidIngesterException:
            # Constructors fail when they cannot find appropriate files
            pass
    if not found_parser:
        raise Exception('Directory is not in correct format for an existing parser')
    if verbose > 0:
        print("Found a {} directory".format(parser.get_name()))
        
    # Get information about the chemical system
    chem = ChemicalSystem()
    chem.chemical_formula = parser.get_composition()
        
    # Get software information, to list as method
    software = Software(name=parser.get_name(),
        version=parser.get_version_number())
        
    # Define the DFT method object
    method = Method(name='Density Functional Theory',
        software=[software])
        
    # Get the settings (aka. "conditions") of the DFT calculations
    conditions = []
    for name, func in parser.get_setting_functions().items():
        # Get the condition
        cond = getattr(parser, func)()

        # If the condition is None or False, skip it
        if cond is None:
            continue

        if inline and cond.files is not None:
            continue

        # Set the name
        cond.name = name

        # Set the types
        conditions.append(cond)
    
    # Get the properties of the system
    chem.properties = []
    for name, func in parser.get_result_functions().items():
        # Get the property
        prop = getattr(parser, func)()
        
        # If the property is None, skip it
        if prop is None:
            continue

        if inline and prop.files is not None:
            continue

        # Add name and other data
        prop.name = name
        prop.methods = [method,]
        prop.data_type='COMPUTATIONAL'
        if verbose > 0 and isinstance(prop, Value):
            print(name)
        if prop.conditions is None:
            prop.conditions = conditions
        else:
            if not isinstance(prop.conditions, list):
                prop.conditions = [prop.conditions]
            prop.conditions.extend(conditions)

        # Add it to the output
        chem.properties.append(prop)

    # Check to see if we should add the quality report
    if quality_report and isinstance(parser, VaspParser):
        _add_quality_report(parser, chem)

    return chem


def directory_to_pif(directory, **kwargs):
    """
    Convert a directory to a pif
    :param directory: Directory to convert to a pif
    :param kwargs: any additional keyword arguments. (See `files_to_pif`)
    :return: the created pif
    """

    # Get the files
    files = [os.path.join(directory, f) for f in os.listdir(directory)
             if os.path.isfile(os.path.join(directory, f))]

    # Run the pif
    return files_to_pif(files, **kwargs)


def convert(files, **kwargs):
    """
    Wrap directory to pif as a dice extension
    :param files: a list of files, which must be non-empty
    :param kwargs: any additional keyword arguments
    :return: the created pif
    """

    if len(files) < 1:
        raise ValueError("Files needs to be a non-empty list")

    if len(files) == 1:
        if os.path.isfile(files[0]):
            return files_to_pif(files, **kwargs)
        else:
            return directory_to_pif(files[0], **kwargs)
    else:
        return files_to_pif([x for x in files if os.path.isfile(x)], **kwargs)



from setuptools import setup, find_packages

setup(
    name='dfttopif',
    version='1.0.0',
    description='Library for parsing Density Functional Theory calculations',
    url='https://github.com/CitrineInformatics/pif-dft',
    install_requires=[
        'ase',
        'pypif>=2.0.1,<3',
        'dftparse>=0.2.1'
    ],
    extras_require={
        'report': ["requests"],
    },
    packages=find_packages(exclude=('tests', 'docs')),
    entry_points={
        'citrine.dice.converter': [
            'dft = dfttopif'
        ]
    }
)
